/*
 * Contains the packet driver for the RTK8019as 10megabit ethernet controller
 * chip.
 *
 * IMPORTANT NOTE: Software emulated pullup of the 9346 is not supported. If
 *                 you need to use this wire for another purpose this driver
 *                 will not currently work without modification.
 *
 * (C) 2002 Softools, Inc.
 * (C) 2002 Compendium Technologies, Inc.
 * All rights reserved.
 *
 * Translated from the Z-World packet driver and used under license.
 */

#include <Rabbit2000.h>
#include <string.h>
#include "RTK8019AS.h"
#include "lwip/stats.h"

#define RX_DEBUG
#define TX_DEBUG

char writing = 0;

// Forward declaration of busy delay
static void delay_uS(unsigned);

/**
 * This routine detects the EEPROM clock pulse from the RTK8019AS.
 */
static void clock9346(void)
{
// HL' alternate register comes in with the port address
#asm
        ; Save contents of registers
        ex      af,af'
        exx     
        ld      a,(_rtkEESK##)
        ld      e,a

h2l:
ioi     ld      a,(hl)
        and     e
        jr      z,h2l
l2h:
ioi     ld      a,(hl)
        and     e
        jr      nz,l2h

        ; Get old registers back
        exx     
        ex      af,af'
#endasm
}

/**
 * Emulate the EEPROM loading sequence required by the RTK8019AS in
 * the absence of a real EEPROM. The EEPROM is absent on the RCM21xx
 * series and also on the micro ethernet RCM2200. It is present on
 * the TCP/IP dev kit board.
 */
static void emulate9346(void)
{
#asm
;
; code to emulate the 93C46
;
        ld      c,1
        ld      hl,(_rtkEEPORT##)
        ld      hl',(_rtkEEPORT##)   ; For later use in clock9346 function

emu_loop:
        ld      a,(_rtkEEDI##)
        ld      e,a

        ;
        ; start bit
        ;
emu_1:
        lcall   $clock9346
ioi     ld      a,(hl)
        and     e
        jr      z,emu_1

        ;
        ; read 10 - read command
        ;
emu_2:
        lcall   $clock9346
ioi     ld      a,(hl)
        and     e
        jr      z,emu_1


emu_3:
        lcall   $clock9346
ioi     ld      a,(hl)
        and     e
        jr      nz,emu_1

        ;
        ; six clocks for address
        ;

        ld      b,6
emu_4:
        lcall   $clock9346
        djnz    emu_4

        ;
        ; send 0
        ;

        ld      de,(_rtkEEDO##)
        xor     a
ioi     ld      (de),a

        ;
        ; send word (16 bits)
        ;
        ld      a,(_rtkLEDs##)
        ld      b,16

emu_5:
        lcall   $clock9346
ioi     ld      (de),a
        rlca
        djnz    emu_5


        dec     c
        jp      p,emu_loop


        xor     a
ioi     ld      (de),a

#endasm

}

#ifdef DEBUG
char regbuf[64];

/**
 * This routine loads and prints out the contents of the RealTek 8019AS
 * registers.
 */
void printRealTekRegisters(void)
{
#asm
	push	ix
	push	iy
	;; if not a Blue Cheese, only one NIC available
	ld	ix, (_rtkWriteAddress)
	ld	iy, (_rtkReadAddress)

ioe	ld	a,(iy)
	push	af
		
	ld	de,_regbuf

	ld	a,0x20

ioe	ld	(ix),a
	ld	b,16
	ld	hl, iy
loop1:
ioe	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	loop1
		
	ld	a,0x60
ioe	ld	(ix),a
	ld	b,16
	ld	hl,iy
loop2:
ioe	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	loop2
		
	ld	a,0xa0
ioe	ld	(ix),a
	ld	b,16
	ld	hl,iy
loop3:
ioe	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	loop3
		
	ld	a,0xe0
ioe	ld	(ix),a
	ld	b,16
	ld	hl,iy
loop4:
ioe	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	djnz	loop4
		
	pop	af
	ld	(_regbuf),a

	pop	iy
	pop	ix
#endasm

	printf("CR=%02x\n",regbuf[0]);
	printf("CLDA0=%02x\tPAR0=%02x\t\tPSTART=%02x\t9346CR=%02x\n",
		regbuf[1],regbuf[17],regbuf[33],regbuf[49]);
	printf("CLDA1=%02x\tPAR1=%02x\t\tPSTOP=%02x\tBPAGE=%02x\n",
		regbuf[2],regbuf[18],regbuf[34],regbuf[50]);
	printf("BNRY=%02x\t\tPAR2=%02x\t\t\t\tCONFIG0=%02x\n",
		regbuf[3],regbuf[19],regbuf[51]);
	printf("TSR=%02x\t\tPAR3=%02x\t\tTPSR=%02x\t\tCONFIG1=%02x\n",
		regbuf[4],regbuf[20],regbuf[36],regbuf[52]);
	printf("NCR=%02x\t\tPAR4=%02x\t\t\t\tCONFIG2=%02x\n",
		regbuf[5],regbuf[21],regbuf[53]);
	printf("FIFO=%02x\t\tPAR5=%02x\t\t\t\tCONFIG3=%02x\n",
		regbuf[6],regbuf[22],regbuf[54]);
	printf("ISR=%02x\t\tCURR=%02x\t\t\t\tTEST1=%02x\n",
		regbuf[7],regbuf[23],regbuf[55]);
	printf("CRDA0=%02x\tMAR0=%02x\t\t\t\tCSNSAV=%02x\n",
		regbuf[8],regbuf[24],regbuf[56]);
	printf("CRDA1=%02x\tMAR1=%02x\n",
		regbuf[9],regbuf[25]);
	printf("8019ID0=%02x\tMAR2=%02x\t\t\t\tTEST2=%02x\n",
		regbuf[10],regbuf[26],regbuf[58]);
	printf("8019ID1=%02x\tMAR3=%02x\t\t\t\tINTR=%02x\n",
		regbuf[11],regbuf[27],regbuf[59]);
	printf("RSR=%02x\t\tMAR4=%02x\t\tRCR=%02x\n",
		regbuf[12],regbuf[28],regbuf[44]);
	printf("CNTR0=%02x\tMAR5=%02x\t\tTCR=%02x\n",
		regbuf[13],regbuf[29],regbuf[45]);
	printf("CNTR1=%02x\tMAR6=%02x\t\tDCR=%02x\n",
		regbuf[14],regbuf[30],regbuf[46]);
	printf("CNTR2=%02x\tMAR7=%02x\t\tIMR=%02x\n\n",
		regbuf[16],regbuf[31],regbuf[47]);
}
#endif /* DEBUG */

/**
 * Initialize the rtk8019as, resetting the interface and getting the ethernet
 * address.
 */
void RealTekInit(struct netif * interface)
{
    // Set up the Selects and I/O Strobes
    int macOr;
    u8_t value;

#ifdef DEBUG
    printf("in RTK8019as init\n");
#endif

    if (rtkBoardType == RTK_BOARD_TYPE_RCM2200)
    {
        // RCM2200
        ioSrSetBitsI( PEFR, 0x44 );
        ioSrSetBitsI( IB2CR, 0x90 );
        ioSrSetBitsI( IB6CR, 0xA8 );
        ioSrSetBitsI( PEDDR, 0x4C );
        ioSrClearBitsI( PDDDR, 0x02 );
        ioSrClearBitsI( PDDCR, 0x02 );
        ioSrClearBitsI( PDFR, 0x02 );
        ioSetBitI( PCDR, 4 );    // Start the reset cycle
        delay( 1000 );           // Wait 1000ms for reset to complete
        ioClearBitI( PCDR, 4 );  // End the reset
    }
    else if (rtkBoardType == RTK_BOARD_TYPE_RCM2100)
    {
        // RCM2100
        ioSrSetBitsI( PEDDR, 0x44 );
        ioSrSetBitsI( PEFR, 0x44 );
        ioSrSetBitsI( IB2CR, 0x90 );
        ioSrSetBitsI( IB6CR, 0xA8 );
        ioSrSetBitsI( PDDDR, 0x90 );
        ioSrClearBitsI( PDDCR, 0x02 );
        ioSrClearBitsI( PDFR, 0x02 );
        ioSrSetBitI( PDDR, 4 );   // Start the reset cycle
        delay( 1000 );            // Wait 1000ms for reset to complete
        ioSrClearBitI( PDDR, 4 ); // End the reset
    }
    else
    {
        // Assume TCP/IP dev kit
        oute( PEB1R, 0xFF );
        ioSrSetBitsI( PEFR, 0xC0 );     //      7,6 io strobes
        ioSrSetBitsI( IB7CR, 0x90 );    //      Set the io strobes
        ioSrSetBitsI( IB6CR, 0xA8 );
        ioSrSetBitsI( PEDDR, 0xC2 );
        ioClearBitsI( PEDR, 0xFD );
        oute( PEB1R, 0xFD );        //      Enable RealTek
        //  Wait for NIC to reset and establish link...
        delay( 2000 );
    }

    // Reset loop
    while(ine(rtkReadAddress) == 0x7E);
    // No Green Link LED without this delay here on emulated EEPROM boards
    delay(10);     
#ifdef DEBUG
    puts("past reset loop\n");
#endif

    //  Read then write to the RESET register

    value = ine( rtkReadAddress+PD_RESET );
    delay_uS( 1600 );                              //  Wait 1.6ms
    oute( rtkWriteAddress+PD_RESET, value );           //  Reset NIC
    // Stop the NIC, Abort DMA, page 0
    oute( rtkWriteAddress, 0x21 );
    // Wait at least 1.6 milliseconds
    delay_uS( 1600 );

    //  Poll the reset bit in the INTSTATUS register
    //  Clear out all interrupts
    oute( rtkWriteAddress+PD_INTSTATUS, 0xFF );    //  Clear all sticky int bits

    if(rtkBoardType == RTK_BOARD_TYPE_DEVKIT)
    {

        oute( rtkWriteAddress+PD_INTMASK, 0 );         //  Mask out ALL ints

        //  Read contents of the 93C46 into EthAddress
        //  and copy the physical address into the NICs
        //  addresses.  This shouldn't be necessary, but
        //  the 93C46 wasn't being properly loaded on bootup

        //  12 bytes from address 0
        oute( rtkWriteAddress+PD_RMTBCNT0, 0x12 );

        //  Setup dma address/len
        oute( rtkWriteAddress+PD_RMTBCNT1, 0 );
        oute( rtkWriteAddress+PD_RMTSADR0, 0 );
        oute( rtkWriteAddress+PD_RMTSADR1, 0 );

        oute( rtkWriteAddress, 0x0A );             //  Start dma read

        //  Read six bytes into EthAddress
        for( char i = 0; i < (char) sizeof interface->hwaddr; ++i )
        {
            ine( PD_IOPORT+rtkReadAddress );
            interface->hwaddr[i] = ine( PD_IOPORT+rtkReadAddress );
        }

    }
    else
    {
        struct ethernetif *ethif;
        // Core module ethernets do not have an EEPROM, therefore
        // emulation must be done to lines attached to the 8019as to act
        // as if the Rabbit 2000 is the EEPROM.

        ipset2();

        // Switch Realtek registers to page three
        oute(rtkWriteAddress, 0xC0);

        // Trigger the 9346 emulation. This code will play the part of the 9346
        // EEPROM that isn't here
        oute(rtkWriteAddress+PD_9346CR, 0x40);

        // 9346 emulation code here
        emulate9346();
        delay(10);
#ifdef DEBUG
        puts("ending emulation code\n");
#endif
        ipres();
        delay_uS( 500 );         //  Wait 0.5ms

        // Assign ethernet address
        ethif = interface->state;
        *(ethif->ethaddr) = rtkEthernetAddress;
    }

    // Do I have an ethernet address? Complain if I don't
    macOr = 0;
    for (char i = 0; i < (char) sizeof interface->hwaddr; i++)
    {
        macOr |= interface->hwaddr[i];
    }


    if (!macOr)
    {
        puts("No MAC address (ethernet address) found or given\n");
        return;
    }


    // This device initialization is the same for all boards / modules. The only
    // thing that differs is the page addresses.
    // Stop nic, change to register page 1
    oute( rtkWriteAddress, 0x60 );


    //  Write the ethernet address (six bytes) into the 8019as registers
    for( unsigned i = 0; i < sizeof interface->hwaddr; ++i )
        oute( rtkWriteAddress+1+i, interface->hwaddr[i] );


    //  initialize MARx to 0 (ie: reject all multicast)
    for( unsigned i = 0; i < 8; ++i )
        oute( rtkWriteAddress+8+i, 0 );


    //  Initialize the interface


    //  Stop nic, abort remote dma, reset to register page 0
    oute( rtkWriteAddress, 0x21 );


    delay_uS( 1600 );                              //  Wait 1.6ms


    //  Normal operation, initialize remote Dma, fifo threshhold 8 bytes
    oute( rtkWriteAddress+PD_DATACFG, 0x58 );


    //  Remote dma byte count = 0000h
    oute( rtkWriteAddress+PD_RMTBCNT0, 0 );
    oute( rtkWriteAddress+PD_RMTBCNT1, 0 );
    oute( rtkWriteAddress+PD_RMTSADR0, 0 );
    //  Remote dma start address = 4000h
    oute( rtkWriteAddress+PD_RMTSADR1, 0x40 );
    oute( rtkWriteAddress+PD_RXCFG, 0x20 );        //  Monitor mode
    oute( rtkWriteAddress+PD_TXCFG, 0x02 );        //  Place NIC in loopback


    //  Tx buffer < 4600h <= rx buffer
    oute( rtkWriteAddress+PD_PAGESTART, 0x46 );
    oute( rtkWriteAddress+PD_BOUNDARY, 0x46 );


    //  Rx buffer < 8000h
    oute( rtkWriteAddress+PD_PAGESTOP, PD_PAGE_STOP );


    //  Clear all interrupt flags
    oute( rtkWriteAddress+PD_INTSTATUS, 0xFF );


    //  Disable interrupt generation
    //oute( rtkWriteAddress+PD_INTMASK, 0 );


    oute( rtkWriteAddress+PD_TXPAGE, 0x40 );       //  4000h < tx buffer


    //  Stop nic, change to register page 1
    oute( rtkWriteAddress, 0x61 );


    //  Next place to rx a packet
    oute( rtkWriteAddress+PD_CURRENT, 0x46 );


    //  Start nic, abort remote dma
    oute( rtkWriteAddress, 0x22 );


    //  Change from loopback mode to normal op
    oute( rtkWriteAddress+PD_TXCFG, 0 );


    oute( rtkWriteAddress+PD_RXCFG, 0x04 );        //  Accept broadcast packets


    //  Clear any pending interrupts
    oute( rtkWriteAddress+PD_INTSTATUS, 0xFF );


#ifdef DEBUG
    {
        char eth[20];
        printf("MAC address: %s\n",
               etherAddressToString(eth, interface->hwaddr));
        printf("read: 0x%x write: 0x%x\n", rtkReadAddress, rtkWriteAddress);
    }
    printRealTekRegisters();
    puts("leaving RTK8019as init\n");
#endif
}


/**
 * Send a packet to the RTK8019as from a series of pbuf buffers.
 */
int RealTekSendPacket(struct pbuf *p)
{
    struct pbuf * q;
    unsigned int packetLength = p->tot_len;
    unsigned char padLength = 0;

    // Weakness: multiple threads could write at the same time
    ipset3();
    writing = 1;
    ipres();

    if (packetLength < 60)
    {
        padLength = 60 - packetLength;
        packetLength = 60;
    }
#ifdef TX_DEBUG
    printf("<send: %d pad: %d\n", packetLength, padLength);
#endif

    // Semaphore is necessary here or before we get here for lwIP cases.
    // Abort any dma currently active
    oute( rtkWriteAddress, 0x20 );

    // Wait for completion transmitting prev packet - Bit 2 set if busy
    while( (char) ( ine( rtkReadAddress ) & (1 << 2) ) != 0 );

    // Amount to send
    oute( rtkWriteAddress+PD_RMTBCNT0, packetLength );
    oute( rtkWriteAddress+PD_RMTBCNT1, packetLength >> 8 );

    // Address on NIC to store
    oute( rtkWriteAddress+PD_RMTSADR0, 0 );
    oute( rtkWriteAddress+PD_RMTSADR1, 0x40 );
    oute( rtkWriteAddress, 0x12 );                  // Write command to start

    // Send the packet from the chain of buffers, one byte at a time
    for (q = p; q != NULL; q = q->next)
    {
        unsigned char *payload = q->payload;
        unsigned int i;
        unsigned char j;

        //for (int i = 0; i < q->len; i++)
        for (i = q->len; i; i--)
        {
            //oute(rtkWriteAddress + PD_IOPORT, payload[i]);
#ifdef TX_DEBUG
            //printf("%02x", *payload);
#endif
            oute(rtkWriteAddress + PD_IOPORT, *payload++);
        }

        // Write padding for undersized packets
        for (j = padLength; j; j--)
        {
            oute(rtkWriteAddress + PD_IOPORT, 0);
        }
    }

    //  Wait for dma to complete - Bit 6 clear if busy
    while( (char) ( ine( rtkReadAddress+PD_INTSTATUS ) & (1 << 6) ) == 0 );

    //  Clear dma bit in register
    oute( rtkWriteAddress+PD_INTSTATUS, 0x40 );
    //  Addr in ring to transmit
    oute( rtkWriteAddress+PD_TXPAGE, 0x40 );
    oute( rtkWriteAddress+PD_TXBCNT0, packetLength );    //      Amount to send
    oute( rtkWriteAddress+PD_TXBCNT1, packetLength >> 8 );

    //  Start transmission (and shut off remote dma)
    oute( rtkWriteAddress, 0x26 );

#ifdef LINK_STATS
    stats.link.xmit++;
#endif /* LINK_STATS */      
#ifdef TX_DEBUG
    //printf("done transmitting packet\n");
#endif

    ipset3();
    writing = 0;
    ipres();
    return ERR_OK;
}


/**
 * Read a packet into a pbuf chain.
 */
static struct pbuf * readPacket(void)
{
    unsigned int  packetLength;
    unsigned char PDHeader[18];   // Temp storage for DMA and Ethernet headers
    struct pbuf * p;
    struct pbuf * q;

    //  Read header
    oute( rtkWriteAddress, 0x1A );           //  Start dma next packet
    for( unsigned char i = 0; i < 18; i++ )
    {
        PDHeader[i] = ine( PD_IOPORT+rtkReadAddress );
    }


    //  Check for a good packet - Bit 0 set checking all but bit 5
    if( ( PDHeader[0] & 0xDF ) != 1 )
    {
        char current;
        char boundary;
        boundary = ine( PD_BOUNDARY+rtkReadAddress );
        if( boundary <= 0x45 || boundary > PD_PAGE_STOP)
            //  Header bad, reset to 0x46
            oute( rtkWriteAddress+PD_BOUNDARY, current = 0x46 );
        else
            //  Set boundary
            current = PD_PAGE_STOP+1;

        oute( rtkWriteAddress, 0x62 );           //  Page 1 + abort dma
        oute( rtkWriteAddress+PD_CURRENT, current );
        oute( rtkWriteAddress, 0x22 );           //  Page 0 + abort dma
        return NULL;
    }


    //  Check for IP or ARP exception
    if( PDHeader[16] != 8 || ( PDHeader[17] != 0 && PDHeader[17] != 6 ) )
    {
        oute( rtkWriteAddress, 0x22 );           //  Page 0 + abort dma
        oute( rtkWriteAddress+PD_BOUNDARY, PDHeader[1] );   //  Delete packet
        return NULL;
    }

    //  Store real length, set len to packet length - header
    packetLength = ((unsigned) PDHeader[2] | ( PDHeader[3] << 8 ));
#ifdef RX_DEBUG
    //puts(">");
    printf("packet length: %d\n", packetLength);
#endif

    // Allocate the right amount for the packet from the memory pool
    p = pbuf_alloc(PBUF_LINK, packetLength, PBUF_POOL);
  
    if (p != NULL) {

        // This assumes a minimum pbuf size of 14 ... a good assumption
        memcpy(p->payload, PDHeader + 4, 14);

        // Iterate over the pbuf chain, reading the entire packet minus
        // the header into storage.
        for (q = p; q != NULL; q = q->next) {

            unsigned char * payload = q->payload;
            unsigned char * payload_end = payload + q->len;

            // First 14 bytes are already there, skip them
            if (q == p)
            {
                payload += 14;
            }

            // Fill this pbuf byte by byte in the chain.
            while (payload != payload_end)
            {
                *payload++ = ine(PD_IOPORT + rtkReadAddress);
            }
        }
#ifdef LINK_STATS
        stats.link.recv++;
#endif /* LINK_STATS */      
    }
    else
    {
        // Read remainder of packet and discard

        for(unsigned int i = packetLength - 14; i; --i)
        {
            ine(PD_IOPORT + rtkReadAddress);
        }
#ifdef LINK_STATS
        stats.link.memerr++;
        stats.link.drop++;
#endif /* LINK_STATS */      
    }

    return p;  
}


/**
 * Read a packet, clearing overflows.
 */
struct pbuf * RealTekRecvPacket()
{
    char current;
    char boundary;

    // Don't try to read if writing is in progress.
    ipset3();
    if (writing)
    {
        ipres();
        return NULL;
    }
    ipres();

#ifdef RX_DEBUG
    //printf("*");
#endif
    oute( rtkWriteAddress, 0x20 );          //  Select page 0

    if ( ( (char) ine( rtkReadAddress+PD_INTSTATUS ) & (char) (1 << 4) ) != 0 )
    {
        pbuf_free(readPacket());
        oute( rtkReadAddress+PD_INTSTATUS, 0x10 );  //  Clear overflow
        return NULL;
    }

    oute( rtkWriteAddress, 0x62 );          //  Page 1 + abort dma
    current = ine( rtkReadAddress+PD_CURRENT );
    oute( rtkWriteAddress, 0x22 );          //  Page 0 + abort dma
    boundary = ine( rtkReadAddress+PD_BOUNDARY );

    //  Check for no packet
    if (current == boundary)
    {
        //return E_Interface_NoPacket;
        return NULL;
    }

    return readPacket();
}


//      Delay microseconds.
//      Uses the RTC clock - waiting for 488uS periods.


void delay_uS( unsigned uS )
{
    unsigned base;

    if( uS == 0 )
        return;
    outi( RTC0R, 0 );                   //      Latch clock
    //  Count of clock to wait (wrap is ok here)
    base = ((ini( RTC1R ) << 8) | (unsigned) ini( RTC0R )) + ((uS + 487) / 488);
    do  {
        outi( RTC0R, 0 );
    }
    while( ((ini( RTC1R ) << 8) | (unsigned) ini( RTC0R )) < base );
}
